GCField("Function",GCFunction,                            function,SvGCint)             /* logical operation */
GCField("PlaneMask",GCPlaneMask,                          plane_mask,SvGCunsigned_long) /* plane mask */
GCField("Foreground",GCForeground,                        foreground,SvGCunsigned_long) /* foreground pixel */
GCField("Background",GCBackground,                        background,SvGCunsigned_long) /* background pixel */
GCField("LineWidth",GCLineWidth,                          line_width,SvGCint)           /* line width */
GCField("LineStyle",GCLineStyle,                          line_style,SvGCint)           /* LineSolid, LineOnOffDash, LineDoubleDash */
GCField("CapStyle",GCCapStyle,                            cap_style,SvGCint)            /* CapNotLast, CapButt, CapRound, CapProjecting */
GCField("JoinStyle",GCJoinStyle,                          join_style,SvGCint)           /* JoinMiter, JoinRound, JoinBevel */
GCField("FillStyle",GCFillStyle,                          fill_style,SvGCint)           /* FillSolid, FillTiled, FillStippled, FillOpaeueStippled */
GCField("FillRule",GCFillRule,                            fill_rule,SvGCint)            /* EvenOddRule, WindingRule */
GCField("Tile",GCTile,                                    tile,SvGCPixmap)              /* tile pixmap for tiling operations */
GCField("Stipple",GCStipple,                              stipple,SvGCPixmap)           /* stipple 1 plane pixmap for stipping */
GCField("TileStipXOrigin",GCTileStipXOrigin,              ts_x_origin,SvGCint)          /* offset for tile or stipple operations */
GCField("TileStipYOrigin",GCTileStipYOrigin,              ts_y_origin,SvGCint)
GCField("Font",GCFont,                                    font,SvGCFont)                /* default text font for text operations */
GCField("SubwindowMode",GCSubwindowMode,                  subwindow_mode,SvGCint)       /* ClipByChildren, IncludeInferiors */
GCField("GraphicsExposures",GCGraphicsExposures,          graphics_exposures,SvGCBool)  /* boolean, should exposures be generated */
GCField("ClipXOrigin",GCClipXOrigin,                      clip_x_origin,SvGCint)        /* origin for clipping */
GCField("ClipYOrigin",GCClipYOrigin,                      clip_y_origin,SvGCint)
GCField("ClipMask",GCClipMask,                            clip_mask,SvGCPixmap)         /* bitmap clipping; other calls for rects */
GCField("ArcMode",GCArcMode,                              arc_mode,SvGCint)             /* ArcChord, ArcPieSlice */


























